# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The implementation of the  GenericType and GenericMeta class below is
inspired by the implementation by the Tensorflow Extended team, which can be
found here:

https://github.com/tensorflow/tfx/blob/master/tfx/dsl/component/experimental/annotations.py

The list of changes to improve the interaction with ZenML:

- The main classes are merged into a single class which serves as the base
class to be extended
- A class variable is added to be utilized while checking the type of T
- A few more classes are created on top of artifacts and primitives in order to
use the same paradigm with steps and datasources as well

"""
import inspect
from typing import Type

from six import with_metaclass


class BaseAnnotationMeta(type):
    """The Metaclass for the annotations in ZenML. It defines a __get_item__
    method which in return class the __generic_getitem of the main class
    with the same param
    """

    def __getitem__(cls: Type["BaseAnnotation"], params):
        return cls._generic_getitem(params)


class BaseAnnotation(with_metaclass(BaseAnnotationMeta, object)):
    """A main generic class which will be used as the base for annotations"""

    VALID_TYPES = None

    def __init__(self, object_type, _init_via_getitem=False):
        """Initialization of the BaseAnnotation

        The usage of this initialization is only allowed through the
        metaclass.
        """
        if not _init_via_getitem:
            class_name = self.__class__.__name__
            raise ValueError(
                f"{class_name} should be instantiated via %s[T], "
                f"where T is one of {self.VALID_TYPES}"
            )
        self.type = object_type

    def __repr__(self):
        """Representation of the annotation object"""
        return "%s[%s]" % (self.__class__.__name__, self.type)

    @classmethod
    def _generic_getitem(cls, params):
        """
        A getitem method which is used by the metaclass upon class definition

        Args:
          params: cls, a type of a class which is given with the annotation

        Returns:
            an instance of the annotation with the right object type
        """

        # Check the type of the given param and fail if it is not a valid type
        if inspect.isclass(params) and any(
            issubclass(params, t) for t in cls.VALID_TYPES
        ):
            return cls(params, _init_via_getitem=True)
        else:
            class_name = cls.__name__
            raise ValueError(
                f"Generic type `{class_name}[T]` expects the "
                f"single parameter T to be one of "
                f"{cls.VALID_TYPES}."
            )



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import logging
import os
import sys
from logging.handlers import TimedRotatingFileHandler
from typing import Any

from absl import logging as absl_logging

from zenml.constants import (  # isort: skip
    ABSL_LOGGING_VERBOSITY,
    APP_NAME,
    ZENML_LOGGING_VERBOSITY,
)

# Mute tensorflow cuda warnings
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

LOGGING_LEVELS = {
    0: logging.NOTSET,
    1: logging.ERROR,
    2: logging.WARN,
    3: logging.INFO,
    4: logging.DEBUG,
}

FORMATTING_STRING = "%(asctime)s — %(name)s — %(levelname)s — %(message)s"
FORMATTER = logging.Formatter(FORMATTING_STRING)
LOG_FILE = f"{APP_NAME}_logs.log"


def resolve_logging_level():
    """Resolve the logging level from the env variable."""
    if ZENML_LOGGING_VERBOSITY > 0:
        level = (
            LOGGING_LEVELS[ZENML_LOGGING_VERBOSITY]
            if ZENML_LOGGING_VERBOSITY in LOGGING_LEVELS
            else logging.DEBUG
        )
    else:
        level = logging.NOTSET
    return level


def set_root_verbosity():
    """Set the root verbosity."""
    level = resolve_logging_level()
    if level > logging.NOTSET:
        logging.basicConfig(level=level)
        get_logger(__name__).debug(
            f"Logging set to level: " f"{logging.getLevelName(level)}"
        )
    else:
        logging.disable(sys.maxsize)
        logging.getLogger().disabled = True
        get_logger(__name__).debug("Logging NOTSET")


def get_console_handler() -> Any:
    """Get console handler for logging."""
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler() -> Any:
    """Return a file handler for logging."""
    file_handler = TimedRotatingFileHandler(LOG_FILE, when="midnight")
    file_handler.setFormatter(FORMATTER)
    return file_handler


def get_logger(logger_name) -> Any:
    """Main function to get logger name,.

    Args:
      logger_name: Name of logger to initialize.

    Returns:
        A logger object.

    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(resolve_logging_level())
    logger.addHandler(get_console_handler())

    # TODO: [LOW] Add a file handler for persistent handling
    # logger.addHandler(get_file_handler())
    # with this pattern, it's rarely necessary to propagate the error up to
    # parent
    logger.propagate = False
    return logger


def init_logging():
    """Initialize logging with default levels."""
    logging.basicConfig(format=FORMATTING_STRING)
    set_root_verbosity()

    # set absl logging
    absl_logging.set_verbosity(ABSL_LOGGING_VERBOSITY)



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import random

from zenml.enums import PipelineStatusTypes
from zenml.pipelines.training_pipeline import TrainingPipeline

# we expect all queries to fail since the metadata store
# cannot be instantiated
expected_query_error = AssertionError


def test_metadata_init(repo):
    """

    Args:
      repo:

    Returns:

    """
    mds = repo.get_default_metadata_store()
    _ = mds.store


def test_to_config(repo):
    """

    Args:
      repo:

    Returns:

    """
    mds = repo.get_default_metadata_store()
    mds.to_config()


def test_get_pipeline_status(repo):
    """

    Args:
      repo:

    Returns:

    """
    random_pipeline = random.choice(repo.get_pipelines())

    mds = repo.get_default_metadata_store()

    assert (
        mds.get_pipeline_status(random_pipeline)
        == PipelineStatusTypes.Succeeded.name
    )


def test_get_pipeline_executions(repo):
    """

    Args:
      repo:

    Returns:

    """
    mds = repo.get_default_metadata_store()

    random_pipeline = random.choice(repo.get_pipelines())

    _ = mds.get_pipeline_executions(random_pipeline)


def test_get_components_status(repo):
    """

    Args:
      repo:

    Returns:

    """
    mds = repo.get_default_metadata_store()

    random_pipeline = random.choice(repo.get_pipelines())

    mds.get_components_status(random_pipeline)


def test_get_artifacts_by_component(repo):
    """

    Args:
      repo:

    Returns:

    """
    mds = repo.get_default_metadata_store()

    random_pipeline = repo.get_pipelines_by_type(
        TrainingPipeline.PIPELINE_TYPE
    )[0]

    # pick a component guaranteed to be present
    component_name = ""  # GDPComponent.SplitGen.name

    artifacts = mds.get_artifacts_by_component(random_pipeline, component_name)

    assert len(artifacts) >= 1


def test_get_pipeline_context(repo):
    """

    Args:
      repo:

    Returns:

    """
    mds = repo.get_default_metadata_store()

    random_pipeline = random.choice(repo.get_pipelines())

    mds.get_pipeline_context(random_pipeline)



import os
from abc import abstractmethod
from typing import Text

from zenml.core.base_component import BaseComponent
from zenml.core.component_factory import orchestrator_store_factory
from zenml.enums import OrchestratorTypes
from zenml.utils.path_utils import get_zenml_config_dir


@orchestrator_store_factory.register(OrchestratorTypes.base)
class BaseOrchestrator(BaseComponent):
    _ORCHESTRATOR_STORE_DIR_NAME: Text = "orchestrators"

    @abstractmethod
    def run(self, zenml_pipeline, **kwargs):
        pass

    def get_serialization_dir(self):
        """Gets the local path where artifacts are stored."""
        return os.path.join(
            get_zenml_config_dir(), self._ORCHESTRATOR_STORE_DIR_NAME
        )

    class Config:
        """Configuration of settings."""

        env_prefix = "zenml_orchestrator_"



# #  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
# #
# #  Licensed under the Apache License, Version 2.0 (the "License");
# #  you may not use this file except in compliance with the License.
# #  You may obtain a copy of the License at:
# #
# #       http://www.apache.org/licenses/LICENSE-2.0
# #
# #  Unless required by applicable law or agreed to in writing, software
# #  distributed under the License is distributed on an "AS IS" BASIS,
# #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# #  or implied. See the License for the specific language governing
# #  permissions and limitations under the License.
# """CLI for pipelines."""
#
# from typing import Text
#
# import click
# from tabulate import tabulate
#
# from zenml.cli.cli import cli
# from zenml.cli.utils import error, pretty_print, pass_repo
# from zenml.pipelines import TrainingPipeline
# from zenml.core import Repository
# from zenml.utils.yaml_utils import read_yaml
#
#
# @cli.group()
# def pipeline():
#     """Pipeline group"""
#     pass
#
#
# @pipeline.command('compare')
# @pass_repo
# def compare_training_runs(repo: Repository):
#     """Compares pipelines in repo"""
#     click.echo('Comparing training pipelines in repo: Starting app..')
#     repo.compare_training_runs()
#
#
# @pipeline.command('list')
# @pass_repo
# def list_pipelines(repo: Repository):
#     """Lists pipelines in the current repository."""
#     try:
#         pipelines = repo.get_pipelines()
#
#         names = [p.name for p in pipelines]
#         types = [p.PIPELINE_TYPE for p in pipelines]
#         statuses = [p.get_status() for p in pipelines]
#         cache_enabled = [p.enable_cache for p in pipelines]
#         filenames = [p.file_name for p in pipelines]
#
#         headers = ["name", "type", "cache enabled", "status", "file name"]
#
#         click.echo(tabulate(zip(names, types, cache_enabled,
#                                 statuses, filenames),
#                             headers=headers))
#     except Exception as e:
#         error(e)
#
#
# @pipeline.command('get')
# @click.argument('pipeline_name')
# @pass_repo
# def get_pipeline_by_name(repo: Repository, pipeline_name: Text):
#     """
#     Gets pipeline from current repository by matching a name against a
#     pipeline name in the repository.
#     """
#     try:
#         p = repo.get_pipeline_by_name(pipeline_name)
#     except Exception as e:
#         error(e)
#         return
#
#     pretty_print(p)
#
#
# @pipeline.command('run')
# @click.argument('path_to_config')
# def run_pipeline(path_to_config: Text):
#     """
#     Runs pipeline specified by the given config YAML object.
#
#     Args:
#         path_to_config: Path to config of the designated pipeline.
#          Has to be matching the YAML file name.
#     """
#     # config has metadata store, backends and artifact store,
#     # so no need to specify them
#     print(path_to_config)
#     try:
#         config = read_yaml(path_to_config)
#         p: TrainingPipeline = TrainingPipeline.from_config(config)
#         p.run()
#     except Exception as e:
#         error(e)



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""CLI for manipulating ZenML local and global config file."""

from typing import List, Text

import click

from zenml.cli import utils as cli_utils
from zenml.cli.cli import cli
from zenml.config.global_config import GlobalConfig
from zenml.core.component_factory import (
    artifact_store_factory,
    metadata_store_factory,
    orchestrator_store_factory,
)
from zenml.core.repo import Repository


# Analytics
@cli.group()
def analytics():
    """Analytics for opt-in and opt-out"""


@analytics.command(
    "opt-in", context_settings=dict(ignore_unknown_options=True)
)
def opt_in():
    """Opt-in to analytics"""
    GlobalConfig().analytics_opt_in = True
    cli_utils.declare("Opted in to analytics.")


@analytics.command(
    "opt-out", context_settings=dict(ignore_unknown_options=True)
)
def opt_out():
    """Opt-out to analytics"""
    GlobalConfig().analytics_opt_in = False
    cli_utils.declare("Opted in to analytics.")


# Metadata Store
@cli.group()
def metadata():
    """Utilities for metadata store"""


@metadata.command(
    "register", context_settings=dict(ignore_unknown_options=True)
)
@click.argument("metadata_store_name", type=str)
@click.argument("metadata_store_type", type=str)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def register_metadata_store(
    metadata_store_name: Text, metadata_store_type: Text, args: List[Text]
):
    """Register a metadata store."""

    try:
        parsed_args = cli_utils.parse_unknown_options(args)
    except AssertionError as e:
        cli_utils.error(str(e))
        return

    repo: Repository = Repository()
    try:
        comp = metadata_store_factory.get_single_component(metadata_store_type)
    except AssertionError as e:
        cli_utils.error(str(e))
        return

    metadata_store = comp(**parsed_args)
    service = repo.get_service()
    service.register_metadata_store(metadata_store_name, metadata_store)


@metadata.command("list")
def list_metadata_stores():
    """List all available metadata stores from service."""
    service = Repository().get_service()
    cli_utils.title("Metadata Stores:")
    cli_utils.echo_component_list(service.metadata_stores)


@metadata.command("delete")
@click.argument("metadata_store_name", type=str)
def delete_metadata_store(metadata_store_name: Text):
    """Delete a metadata store."""
    service = Repository().get_service()
    cli_utils.declare(f"Deleting metadata store: {metadata_store_name}")
    service.delete_metadata_store(metadata_store_name)
    cli_utils.declare("Deleted!")


# Artifact Store
@cli.group()
def artifact():
    """Utilities for artifact store"""


@artifact.command(
    "register", context_settings=dict(ignore_unknown_options=True)
)
@click.argument("artifact_store_name", type=str)
@click.argument("artifact_store_type", type=str)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def register_artifact_store(
    artifact_store_name: Text, artifact_store_type: Text, args: List[Text]
):
    """Register an artifact store."""

    try:
        parsed_args = cli_utils.parse_unknown_options(args)
    except AssertionError as e:
        cli_utils.error(str(e))
        return

    repo: Repository = Repository()
    comp = artifact_store_factory.get_single_component(artifact_store_type)
    artifact_store = comp(**parsed_args)
    service = repo.get_service()
    service.register_artifact_store(artifact_store_name, artifact_store)


@artifact.command("list")
def list_artifact_stores():
    """List all available artifact stores from service."""
    service = Repository().get_service()
    cli_utils.title("Artifact Stores:")
    cli_utils.echo_component_list(service.artifact_stores)


@artifact.command("delete")
@click.argument("artifact_store_name", type=str)
def delete_artifact_store(artifact_store_name: Text):
    """Delete a artifact store."""
    service = Repository().get_service()
    cli_utils.declare(f"Deleting artifact store: {artifact_store_name}")
    service.delete_artifact_store(artifact_store_name)
    cli_utils.declare("Deleted!")


# Orchestrator
@cli.group()
def orchestrator():
    """Utilities for orchestrator"""


@orchestrator.command(
    "register", context_settings=dict(ignore_unknown_options=True)
)
@click.argument("orchestrator_name", type=str)
@click.argument("orchestrator_type", type=str)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def register_orchestrator(
    orchestrator_name: Text, orchestrator_type: Text, args: List[Text]
):
    """Register a orchestrator."""

    try:
        parsed_args = cli_utils.parse_unknown_options(args)
    except AssertionError as e:
        cli_utils.error(str(e))
        return

    repo: Repository = Repository()
    comp = orchestrator_store_factory.get_single_component(orchestrator_type)
    orchestrator = comp(**parsed_args)
    service = repo.get_service()
    service.register_orchestrator(orchestrator_name, orchestrator)


@orchestrator.command("list")
def list_orchestrators():
    """List all available orchestrators from service."""
    service = Repository().get_service()
    cli_utils.title("Orchestrators:")
    cli_utils.echo_component_list(service.orchestrators)


@orchestrator.command("delete")
@click.argument("orchestrator_name", type=str)
def delete_orchestrator(orchestrator_name: Text):
    """Delete a orchestrator."""
    service = Repository().get_service()
    cli_utils.declare(f"Deleting orchestrator: {orchestrator_name}")
    service.delete_orchestrator(orchestrator_name)
    cli_utils.declare("Deleted!")



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.artifact_stores.local_artifact_store import LocalArtifactStore
from zenml.core.local_service import LocalService
from zenml.metadata.sqlite_metadata_wrapper import SQLiteMetadataStore
from zenml.stacks.base_stack import BaseStack


def test_service_crud():
    """Test basic service crud."""
    # TODO [LOW] [TEST]: Need to improve this logic, potentially with
    #  physically checking the FS us `path_utils`.

    ls = LocalService()
    METADATA_KEY = "metadata local"
    ARTIFACT_KEY = "artifact_local"
    STACK_KEY = "stack_local"
    ORCHESTRATOR_KEY = "orchestrator_local"
    local_artifact_store = LocalArtifactStore()
    local_metadata_store = SQLiteMetadataStore()
    local_stack = BaseStack(
        metadata_store_name=METADATA_KEY,
        artifact_store_name=ARTIFACT_KEY,
        orchestrator_name=ORCHESTRATOR_KEY,
    )

    ls.register_artifact_store(ARTIFACT_KEY, local_artifact_store)
    ls.register_metadata_store(METADATA_KEY, local_metadata_store)
    ls.register_stack(STACK_KEY, local_stack)

    assert ls.get_metadata_store(METADATA_KEY) == local_metadata_store
    assert ls.get_artifact_store(ARTIFACT_KEY) == local_artifact_store
    assert ls.get_stack(STACK_KEY) == local_stack

    ls.delete_artifact_store(ARTIFACT_KEY)
    ls.delete_metadata_store(METADATA_KEY)
    ls.delete_stack(STACK_KEY)

    ls.delete()



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(ROOT_DIR, "VERSION")) as version_file:
    __version__ = version_file.read().strip()


from zenml.logger import init_logging
from zenml.pipelines.pipeline_decorator import pipeline  # noqa
from zenml.steps.step_decorator import step  # noqa

init_logging()



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

""" Plugin which is created to add Google Cloud Store support to ZenML

It inherits from the base Filesystem created by TFX and overwrites the
corresponding functions thanks to gcsfs.

Finally, the plugin is registered in the filesystem registry.
"""

from builtins import FileNotFoundError
from typing import Any, Callable, Iterable, List, Optional, Tuple

import gcsfs
from tfx.dsl.io import filesystem, filesystem_registry
from tfx.dsl.io.filesystem import PathType


class ZenGCS(filesystem.Filesystem):
    SUPPORTED_SCHEMES = ["gs://"]

    fs = gcsfs.GCSFileSystem()

    @staticmethod
    def open(name: PathType, mode: str = "r") -> Any:
        return ZenGCS.fs.open(path=name, mode=mode)

    @staticmethod
    def copy(src: PathType, dst: PathType, overwrite: bool = False) -> None:
        # TODO: additional params
        ZenGCS.fs.copy(path1=src, path2=dst)

    @staticmethod
    def exists(path: PathType) -> bool:
        return ZenGCS.fs.exists(path=path)

    @staticmethod
    def glob(pattern: PathType) -> List[PathType]:
        return ZenGCS.fs.glob(path=pattern)

    @staticmethod
    def isdir(path: PathType) -> bool:
        return ZenGCS.fs.isdir(path=path)

    @staticmethod
    def listdir(path: PathType) -> List[PathType]:
        return ZenGCS.fs.listdir(path=path)

    @staticmethod
    def makedirs(path: PathType) -> None:
        ZenGCS.fs.makedirs(path=path, exist_ok=True)

    @staticmethod
    def mkdir(path: PathType) -> None:
        ZenGCS.fs.makedir(path=path)

    @staticmethod
    def remove(path: PathType) -> None:
        ZenGCS.fs.rm_file(path=path)

    @staticmethod
    def rename(src: PathType, dst: PathType, overwrite: bool = False) -> None:
        # TODO: additional params
        ZenGCS.fs.rename(path1=src, path2=dst)

    @staticmethod
    def rmtree(path: PathType) -> None:
        try:
            ZenGCS.fs.delete(path=path, recursive=True)
        except FileNotFoundError as e:
            raise filesystem.NotFoundError() from e

    @staticmethod
    def stat(path: PathType) -> Any:
        ZenGCS.fs.stat(path=path)

    @staticmethod
    def walk(
        top: PathType,
        topdown: bool = True,
        onerror: Optional[Callable[..., None]] = None,
    ) -> Iterable[Tuple[PathType, List[PathType], List[PathType]]]:
        # TODO: additional params
        return ZenGCS.fs.walk(path=top)


# TODO: [LOW] The registration of the filesystem should happen probably at an
#   artifact store basis
filesystem_registry.DEFAULT_FILESYSTEM_REGISTRY.register(ZenGCS, priority=15)



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


import os
from abc import abstractmethod
from typing import Text

from ml_metadata.metadata_store import metadata_store

from zenml.core.base_component import BaseComponent
from zenml.core.component_factory import metadata_store_factory
from zenml.enums import MLMetadataTypes, PipelineStatusTypes
from zenml.exceptions import DoesNotExistException
from zenml.logger import get_logger
from zenml.utils.path_utils import get_zenml_config_dir

logger = get_logger(__name__)


@metadata_store_factory.register(MLMetadataTypes.base)
class BaseMetadataStore(BaseComponent):
    """Metadata store base class to track metadata of zenml first class
    citizens."""

    _run_type_name: str = "pipeline_run"
    _node_type_name: str = "node"
    _METADATA_STORE_DIR_NAME = "metadata_stores"

    @property
    def store(self):
        """General property that hooks into TFX metadata store."""
        return metadata_store.MetadataStore(self.get_tfx_metadata_config())

    @abstractmethod
    def get_tfx_metadata_config(self):
        """Return tfx metadata config."""

    def get_serialization_dir(self):
        """Gets the local path where artifacts are stored."""
        return os.path.join(
            get_zenml_config_dir(), self._METADATA_STORE_DIR_NAME
        )

    def get_pipeline_status(self, pipeline) -> Text:
        """Query metadata store to find status of pipeline.

        Args:
          pipeline(BasePipeline): a ZenML pipeline object

        Returns:

        """
        try:
            components_status = self.get_components_status(pipeline)
        except Exception:
            return PipelineStatusTypes.NotStarted.name

        for status in components_status.values():
            if status != "complete" and status != "cached":
                return PipelineStatusTypes.Running.name
        return PipelineStatusTypes.Succeeded.name

    def get_pipeline_executions(self, pipeline):
        """Get executions of pipeline.

        Args:
          pipeline(BasePipeline): a ZenML pipeline object

        Returns:

        """
        c = self.get_pipeline_context(pipeline)
        return self.store.get_executions_by_context(c.id)

    def get_components_status(self, pipeline):
        """Returns status of components in pipeline.

        Args:
          pipeline(BasePipeline): a ZenML pipeline object
        Returns: dict of type { component_name : component_status }

        Returns:

        """
        state_mapping = {
            0: "unknown",
            1: "new",
            2: "running",
            3: "complete",
            4: "failed",
            5: "cached",
            6: "cancelled",
        }
        result = {}
        pipeline_executions = self.get_pipeline_executions(pipeline)
        for e in pipeline_executions:
            contexts = self.store.get_contexts_by_execution(e.id)
            node_contexts = [c for c in contexts if c.type_id == 3]
            if node_contexts:
                component_name = node_contexts[0].name.split(".")[-1]
                result[component_name] = state_mapping[e.last_known_state]

        return result

    def get_artifacts_by_component(self, pipeline, component_name: Text):
        """Gets artifacts by component name.

        Args:
          pipeline(BasePipeline): a ZenML pipeline object
          component_name:
          component_name: Text:

        Returns:

        """
        # First get the context of the component and its artifacts
        component_context = [
            c
            for c in self.store.get_contexts_by_type(self._node_type_name)
            if c.name.endswith(component_name)
        ][0]
        component_artifacts = self.store.get_artifacts_by_context(
            component_context.id
        )

        # Second, get the context of the particular pipeline and its artifacts
        pipeline_context = self.store.get_context_by_type_and_name(
            self._run_type_name, pipeline.pipeline_name
        )
        pipeline_artifacts = self.store.get_artifacts_by_context(
            pipeline_context.id
        )

        # Figure out the matching ids
        return [
            a
            for a in component_artifacts
            if a.id in [p.id for p in pipeline_artifacts]
        ]

    def get_pipeline_context(self, pipeline):
        """Get pipeline context.

        Args:
          pipeline:

        Returns:

        """
        # We rebuild context for ml metadata here.
        run_context = self.store.get_context_by_type_and_name(
            type_name=self._run_type_name, context_name=pipeline.pipeline_name
        )
        if run_context is None:
            raise DoesNotExistException(
                name=pipeline.pipeline_name,
                reason="The pipeline does not exist in metadata store "
                "because it has not been run yet. Please run the "
                "pipeline before trying to fetch artifacts.",
            )
        return run_context

    class Config:
        """Configuration of settings."""

        env_prefix = "zenml_metadata_store_"



import inspect
from abc import abstractmethod

from zenml.annotations.artifact_annotations import Input
from zenml.annotations.step_annotations import Step
from zenml.core.repo import Repository
from zenml.utils.exceptions import PipelineInterfaceError


class BasePipelineMeta(type):
    """ """

    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)

        cls.STEP_SPEC = dict()
        cls.INPUT_SPEC = dict()

        connect_spec = inspect.getfullargspec(cls.connect)
        connect_args = connect_spec.args

        if connect_args and connect_args[0] == "self":
            connect_args.pop(0)

        for arg in connect_args:
            arg_type = connect_spec.annotations.get(arg, None)
            if isinstance(arg_type, Step):
                cls.STEP_SPEC.update({arg: arg_type.type})
            elif isinstance(arg_type, Input):
                cls.INPUT_SPEC.update({arg: arg_type.type})
            else:
                raise PipelineInterfaceError(
                    "When you are designing a pipeline, the input signature "
                    "can only parameters which are annotated by either the "
                    "step- and the input annotations."
                )
        return cls


class BasePipeline(metaclass=BasePipelineMeta):
    """ """

    def __init__(self, *args, **kwargs):

        self.__stack = Repository().get_active_stack()

        self.__steps = dict()
        self.__inputs = dict()

        if args:
            raise PipelineInterfaceError(
                "You can only use keyword arguments while you are creating an"
                "instance of a pipeline."
            )

        for k, v in kwargs.items():
            if k in self.STEP_SPEC:
                self.__steps.update({k: v})
            elif k in self.INPUT_SPEC:
                self.__inputs.update({k: v})
            else:
                raise PipelineInterfaceError(
                    f"The argument {k} is an unknown argument. Needs to be "
                    f"one of either {self.INPUT_SPEC.keys()} or "
                    f"{self.STEP_SPEC.keys()}"
                )

    @abstractmethod
    def connect(self, *args, **kwargs):
        """ """

    @classmethod
    def get_executable(cls):
        """ """
        return cls.connect

    @property
    def stack(self):
        return self.__stack

    @stack.setter
    def stack(self, stack):
        raise PipelineInterfaceError(
            "The provider will be automatically"
            "inferred from your environment. Please "
            "do no attempt to manually change it."
        )

    @property
    def inputs(self):
        return self.__inputs

    @inputs.setter
    def inputs(self, inputs):
        raise PipelineInterfaceError(
            "The provider will be automatically"
            "inferred from your environment. Please "
            "do no attempt to manually change it."
        )

    @property
    def steps(self):
        return self.__steps

    @steps.setter
    def steps(self, steps):
        raise PipelineInterfaceError(
            "The provider will be automatically"
            "inferred from your environment. Please "
            "do no attempt to manually change it."
        )

    def run(self):
        return self.stack.orchestrator.run(self)



import inspect
import json
from abc import abstractmethod

from pydantic import create_model

from zenml.annotations import Input, Output
from zenml.artifacts.base_artifact import BaseArtifact
from zenml.artifacts.data_artifact import DataArtifact
from zenml.logger import get_logger
from zenml.steps.utils import generate_component
from zenml.utils.exceptions import StepInterfaceError

logger = get_logger(__name__)


class BaseStepMeta(type):
    """Meta class for `BaseStep`."""

    def __new__(mcs, name, bases, dct):
        """Set up a new class with a qualified spec."""
        logger.debug(f"Registering class {name}, bases: {bases}, dct: {dct}")
        cls = super().__new__(mcs, name, bases, dct)

        cls.INPUT_SPEC = dict()  # all input params
        cls.OUTPUT_SPEC = dict()  # all output params
        cls.PARAM_SPEC = dict()  # all execution params

        # Looking into the signature of the provided process function
        process_spec = inspect.getfullargspec(cls.process)
        process_args = process_spec.args
        logger.debug(f"{name} args: {process_args}")

        # Remove the self from the signature if it exists
        if process_args and process_args[0] == "self":
            process_args.pop(0)

        # Parse the input signature of the function
        for arg in process_args:
            arg_type = process_spec.annotations.get(arg, None)
            if isinstance(arg_type, Input):
                if issubclass(arg_type.type, BaseArtifact):
                    cls.INPUT_SPEC.update({arg: arg_type.type})
                else:
                    cls.INPUT_SPEC.update({arg: DataArtifact})
            elif isinstance(arg_type, Output):
                cls.OUTPUT_SPEC.update({arg: arg_type.type})
            else:
                cls.PARAM_SPEC.update({arg: arg_type})

        # Infer the returned values
        return_spec = process_spec.annotations.get("return", None)
        if return_spec is not None:
            cls.OUTPUT_SPEC.update({"return_output": DataArtifact})

        # Infer the defaults
        cls.PARAM_DEFAULTS = dict()

        process_defaults = process_spec.defaults
        if process_defaults is not None:
            raise StepInterfaceError(
                "The usage of default values for "
                "parameters is not fully implemented yet."
                "Please do not use default values in "
                "your step definition."
            )
            # for i, default in enumerate(process_defaults):
            #     # TODO: [HIGH] fix the implementation
            #     process_args.reverse()
            #     arg = process_args[i]
            #     arg_type = process_spec.annotations.get(arg, None)
            #     if not isinstance(arg_type, Param):
            #         raise StepInterfaceError(
            #             f"A default value in the signature of a step can only "
            #             f"be used for a Param[...] not {arg_type}."
            #         )

        return cls


class BaseStep(metaclass=BaseStepMeta):
    """The base implementation of a ZenML Step which will be inherited by all
    the other step implementations"""

    def __init__(self, *args, **kwargs):
        self.__component_class = generate_component(self)

        # TODO [LOW]: Support args
        if args:
            raise StepInterfaceError(
                "When you are creating an instance of a step, please only "
                "use key-word arguments."
            )

        self.__component = None

        self.__inputs = dict()
        self.__params = dict()

        # TODO: [MEDIUM] add defaults to kwargs
        try:
            # create a pydantic model out of a primitive type
            pydantic_c = create_model(
                "params", **{k: (self.PARAM_SPEC[k], ...) for k in kwargs}
            )
            model_dict = pydantic_c(**kwargs).dict()
            self.__params = {k: json.dumps(v) for k, v in model_dict.items()}

        except RuntimeError:
            # TODO [MED]: Change this to say more clearly what
            #  happened: Even pydantic didnt support this type.
            raise StepInterfaceError()

    @property
    def component(self):
        """Returns a TFX component."""
        if self.__component is None and len(self.INPUT_SPEC) == 0:
            self.__component = generate_component(self)(**self.__params)
        return self.__component

    def __call__(self, **artifacts):
        """Generates a component when called."""
        self.__component = generate_component(self)(
            **artifacts, **self.__params
        )

    def __getattr__(self, item):
        """OVerrides the __getattr__ metho."""
        if item == "outputs":
            return self.component.outputs
        else:
            raise AttributeError(f"{item}")

    @abstractmethod
    def process(self, *args, **kwargs):
        """Abstract method for core step logic."""



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

import pandas as pd

from zenml.materializers.base_materializer import BaseMaterializer
from zenml.utils import path_utils

DEFAULT_FILENAME = "data.csv"


class PandasMaterializer(BaseMaterializer):
    """Materializer to read data to and from pandas."""

    TYPE_NAME = "pandas"

    def read_dataframe(self, filename=None):
        """ """
        filenames = path_utils.list_dir(
            self.artifact.uri, only_file_names=True
        )

        valid_filenames = [
            os.path.join(self.artifact.uri, f)
            for f in filenames
            if DEFAULT_FILENAME in f
        ]

        li = []
        for filename in valid_filenames:
            df = pd.read_csv(filename, index_col=None, header=0)
            li.append(df)
        frame = pd.concat(li, axis=0, ignore_index=True)

        return frame

    def write_dataframe(self, df, filename=None):
        """ """
        filepath = os.path.join(
            self.artifact.uri,
            filename if filename is not None else DEFAULT_FILENAME,
        )

        df.to_csv(filepath)



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

# TODO [MEDIUM]: Add logic via direct GitHub repo pulling

# import os
# from typing import List, Text
#
# import click
#
# from zenml.cli.cli import cli
# from zenml.utils import path_utils
#
#
# def get_examples_dir() -> Text:
#     """Return the examples dir."""
#     cli_dir = os.path.dirname(__file__)
#     package_dir = os.path.dirname(os.path.dirname(cli_dir))
#     examples_dir = os.path.join(package_dir, "examples")
#     return examples_dir
#
#
# def get_all_examples() -> List[Text]:
#     """Get all the examples"""
#     examples = []
#     for name in sorted(os.listdir(get_examples_dir())):
#         # Skip hidden files (like .gitignore)
#         if (
#             not name.startswith(".")
#             and not name.startswith("__")
#             and not name.startswith("README")
#         ):
#             examples.append(name)
#     return examples
#
#
# def get_example_readme(example_path) -> Text:
#     """Get the example README file contents."""
#     with open(os.path.join(example_path, "README.md")) as readme:
#         readme_content = readme.read()
#     return readme_content
#
#
# @cli.group(help="Access all ZenML examples.")
# def example():
#     """Examples group"""
#
#
# @example.command(help="List the available examples.")
# def list():
#     """List all available examples."""
#     click.echo("Listing examples: \n")
#     for name in get_all_examples():
#         click.echo(f"{name}")
#     click.echo("\nTo pull the examples, type: ")
#     click.echo("zenml example pull EXAMPLE_NAME")
#
#
# @example.command(help="Find out more about an example.")
# @click.argument("example_name")
# def info(example_name):
#     """Find out more about an example."""
#     example_dir = os.path.join(get_examples_dir(), example_name)
#     readme_content = get_example_readme(example_dir)
#     click.echo(readme_content)
#
#
# @example.command(
#     help="Pull examples straight " "into your current working directory."
# )
# @click.argument("example_name", required=False, default=None)
# def pull(example_name):
#     """Pull examples straight " "into your current working directory."""
#     examples_dir = get_examples_dir()
#     if not example_name:
#         examples = get_all_examples()
#     else:
#         examples = [example_name]
#
#     # Create destination dir.
#     dst = os.path.join(os.getcwd(), "zenml_examples")
#     path_utils.create_dir_if_not_exists(dst)
#
#     # Pull specified examples.
#     for example in examples:
#         dst_dir = os.path.join(dst, example)
#         # Check if example has already been pulled before.
#         if path_utils.file_exists(dst_dir):
#             if click.confirm(
#                 f"Example {example} is already pulled. "
#                 f"Do you wish to overwrite the directory?"
#             ):
#                 path_utils.rm_dir(dst_dir)
#             else:
#                 continue
#         click.echo(f"Pulling example {example}")
#         src_dir = os.path.join(examples_dir, example)
#         path_utils.copy_dir(src_dir, dst_dir)
#
#         click.echo(f"Example pulled in directory: {dst_dir}")
#
#     click.echo()
#     click.echo(
#         "Please read the README.md file in the respective example "
#         "directory to find out more about the example"
#     )



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text


class MaterializerFactory(object):
    """A factory class which is used by the ZenML artifacts to keep track
    of different materializers"""

    materializer_types = {}

    def __init__(self, artifact):
        """Initialization with an empty factory"""
        self.artifact = artifact

    def __getattr__(self, key: Text):
        """Get a single materializers based on the key.

        Args:
            key: Indicates the materializer name

        Returns:
            The corresponding writer function within the factory.
        """
        if key in self.materializer_types:
            return self.materializer_types[key](self.artifact)
        else:
            return object.__getattribute__(self, key)(self.artifact)

    def get_types(self):
        """Get the materializer dictionary"""
        return self.materializer_types

    @classmethod
    def register_type(cls, key, type_):
        """Register a new materializer in the factory

        Args:
            key: str, which indicates the materializer used within the step

            type_: a ZenML materializer object
        """
        cls.materializer_types[key] = type_



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Factory to register all components."""
from typing import Any, Callable, Dict, Text

from zenml.logger import get_logger

logger = get_logger(__name__)

# TODO [LOW]: Type hints need improving here but need to avoid circular
#  dependencies


class ComponentFactory:
    """Definition of ComponentFactory to track all BaseComponent subclasses.

    All BaseComponents (including custom ones) are to be
    registered here.
    """

    def __init__(self, name: Text):
        """Constructor for the factory.

        Args:
            name: Unique name for the factory.
        """
        self.name = name
        self.components: Dict[Text, Any] = {}

    def get_components(self) -> Dict[Any, Any]:
        """Return all components"""
        return self.components

    def get_single_component(self, key: Text) -> Any:
        """Get a registered component from a key."""
        if key in self.components:
            return self.components[key]
        raise AssertionError(
            f"Type {key} does not exist! Available options: "
            f"{[k for k in self.components.keys()]}"
        )

    def register_component(self, key: Text, component: Any):
        self.components[key] = component

    def register(self, name: str) -> Callable:
        """Class method to register Executor class to the internal registry.

        Args:
            name (str): The name of the executor.

        Returns:
            The Executor class itself.
        """

        def inner_wrapper(wrapped_class: Any) -> Callable:
            """Inner wrapper for decorator."""
            if name in self.components:
                logger.debug(
                    f"Executor {name} already exists for factory {self.name}. Will replace it"
                )
            self.register_component(name, wrapped_class)
            return wrapped_class

        return inner_wrapper


artifact_store_factory = ComponentFactory(name="artifact")
metadata_store_factory = ComponentFactory(name="metadata")
orchestrator_store_factory = ComponentFactory(name="orchestrator")



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


import types
from typing import Text, Type

from zenml.pipelines.base_pipeline import BasePipeline


def pipeline(name: Text = None):
    """Outer decorator function for the creation of a ZenML pipeline

    In order to be able work with parameters such as "name", it features a
    nested decorator structure.

    Args:
        name: str, the given name for the pipeline

    Returns:
        the inner decorator which creates the pipeline class based on the
        ZenML BasePipeline
    """

    def inner_decorator(func: types.FunctionType) -> Type:
        """Inner decorator function for the creation of a ZenML Pipeline

        Args:
          func: types.FunctionType, this function will be used as the
            "connect" method of the generated Pipeline

        Returns:
            the class of a newly generated ZenML Pipeline

        """
        pipeline_class = type(
            name if name else func.__name__,
            (BasePipeline,),
            {"connect": staticmethod(func)},
        )

        return pipeline_class

    return inner_decorator



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Wrapper class to handle Git integration"""
import inspect
import os
from typing import List, Text, Type

from git import BadName
from git import Repo as GitRepo
from tfx.utils.import_utils import import_class_by_path

from zenml.constants import APP_NAME
from zenml.exceptions import GitException
from zenml.logger import get_logger
from zenml.utils import path_utils, source_utils
from zenml.utils.source_utils import (
    get_module_source_from_file_path,
    get_module_source_from_source,
    get_relative_path_from_module_source,
    is_standard_pin,
    is_standard_source,
    resolve_standard_source,
)

GIT_FOLDER_NAME = ".git"

logger = get_logger(__name__)


class GitWrapper:
    """Wrapper class for Git.

    This class is responsible for handling git interactions, primarily
    handling versioning of different steps in pipelines.
    """

    def __init__(self, repo_path: Text):
        """
        Initialize GitWrapper. Should be initialize by ZenML Repository.
        Args:
            repo_path:

        Raises:
            InvalidGitRepositoryError: If repository is not a git repository.
            NoSuchPathError: If the repo_path does not exist.
        """
        # TODO: [LOW] Raise ZenML exceptions here instead.
        self.repo_path: Text = repo_path
        self.git_root_path: Text = os.path.join(repo_path, GIT_FOLDER_NAME)
        self.git_repo: GitRepo = GitRepo(self.repo_path)

    def add_gitignore(self, items: List[Text]):
        """
        Adds `items` to .gitignore, if .gitignore exists. Otherwise creates
        and adds.

        Args:
            items (list[str]): Items to add.
        """
        str_items = "\n".join(items)
        str_items = "\n\n# ZenML\n" + str_items

        gitignore_path = os.path.join(self.repo_path, ".gitignore")
        if not path_utils.file_exists(gitignore_path):
            path_utils.create_file_if_not_exists(gitignore_path, str_items)
        else:
            path_utils.append_file(gitignore_path, str_items)

    def check_file_committed(self, file_path: Text) -> bool:
        """
        Checks file is committed. If yes, return True, else False.

        Args:
            file_path (str): Path to any file within the ZenML repo.
        """
        uncommitted_files = [i.a_path for i in self.git_repo.index.diff(None)]
        try:
            staged_files = [i.a_path for i in self.git_repo.index.diff("HEAD")]
        except BadName:
            # for Ref 'HEAD' did not resolve to an object
            logger.debug("No committed files in the repo. No staged files.")
            staged_files = []

        # source: https://stackoverflow.com/questions/3801321/
        untracked_files = self.git_repo.git.ls_files(
            others=True, exclude_standard=True
        ).split("\n")
        for item in uncommitted_files + staged_files + untracked_files:
            # These are all changed files
            if file_path == item:
                return False
        return True

    def get_current_sha(self) -> Text:
        """
        Finds the git sha that each file within the module is currently on.
        """
        return self.git_repo.head.object.hexsha

    def check_module_clean(self, source: Text):
        """
        Returns True if all files within source's module are committed.

        Args:
            source (str): relative module path pointing to a Class.
        """
        # Get the module path
        module_path = source_utils.get_module_source_from_source(source)

        # Get relative path of module because check_file_committed needs that
        module_dir = source_utils.get_relative_path_from_module_source(
            module_path
        )

        # Get absolute path of module because path_utils.list_dir needs that
        mod_abs_dir = source_utils.get_absolute_path_from_module_source(
            module_path
        )
        module_file_names = path_utils.list_dir(
            mod_abs_dir, only_file_names=True
        )

        # Go through each file in module and see if there are uncommitted ones
        for file_path in module_file_names:
            path = os.path.join(module_dir, file_path)

            # if its .gitignored then continue and dont do anything
            if len(self.git_repo.ignored(path)) > 0:
                continue

            if path_utils.is_dir(os.path.join(mod_abs_dir, file_path)):
                logger.warning(
                    f"The step {source} is contained inside a module "
                    f"that "
                    f"has sub-directories (the sub-directory {file_path} at "
                    f"{mod_abs_dir}). For now, ZenML supports only a flat "
                    f"directory structure in which to place Steps. Please make"
                    f" sure that the Step does not utilize the sub-directory."
                )
            if not self.check_file_committed(path):
                return False
        return True

    def stash(self):
        """Wrapper for git stash"""
        git = self.git_repo.git
        git.stash()

    def stash_pop(self):
        """Wrapper for git stash pop. Only pops if there's something to pop."""
        git = self.git_repo.git
        if git.stash("list") != "":
            git.stash("pop")

    def checkout(self, sha_or_branch: Text = None, directory: Text = None):
        """
        Wrapper for git checkout

        Args:
            sha_or_branch: hex string of len 40 representing git sha OR
            name of branch
            directory (str): relative path to directory to scope checkout
        """
        # TODO: [MEDIUM] Implement exception handling
        git = self.git_repo.git
        if sha_or_branch is None:
            # Checks out directory at sha_or_branch
            assert directory is not None
            git.checkout("--", directory)
        elif directory is not None:
            assert sha_or_branch is not None
            # Basically discards all changes in directory
            git.checkout(sha_or_branch, "--", directory)
        else:
            # The case where sha_or_branch is not None and directory is None
            # In this case, the whole repo is checked out at sha_or_branch
            git.checkout(sha_or_branch)

    def reset(self, directory: Text = None):
        """
        Wrapper for `git reset HEAD <directory>`.

        Args:
            directory (str): relative path to directory to scope checkout
        """
        git = self.git_repo.git
        git.reset("HEAD", directory)

    def resolve_class_source(self, class_source: Text) -> Text:
        """
        Resolves class_source with an optional pin.
        Takes source (e.g. this.module.ClassName), and appends relevant
        sha to it if the files within `module` are all committed. If even one
        file is not committed, then returns `source` unchanged.

        Args:
            class_source (str): class_source e.g. this.module.Class
        """
        if "@" in class_source:
            # already pinned
            return class_source

        if is_standard_source(class_source):
            # that means use standard version
            return resolve_standard_source(class_source)

        # otherwise use Git resolution
        if not self.check_module_clean(class_source):
            # Return the source path if not clean
            logger.warning(
                "Found uncommitted file. Pipelines run with this "
                "configuration may not be reproducible. Please commit "
                "all files in this module and then run the pipeline to "
                "ensure reproducibility."
            )
            return class_source
        return class_source + "@" + self.get_current_sha()

    def is_valid_source(self, source: Text) -> bool:
        """
        Checks whether the source_path is valid or not.

        Args:
            source (str): class_source e.g. this.module.Class[@pin].
        """
        try:
            self.load_source_path_class(source)
        except GitException:
            return False
        return True

    def load_source_path_class(self, source: Text) -> Type:
        """
        Loads a Python class from the source.

        Args:
            source: class_source e.g. this.module.Class[@sha]
        """
        source = source.split("@")[0]
        pin = source.split("@")[-1]
        is_standard = is_standard_pin(pin)

        if "@" in source and not is_standard:
            logger.debug(
                "Pinned step found with git sha. "
                "Loading class from git history."
            )

            module_source = get_module_source_from_source(source)
            relative_module_path = get_relative_path_from_module_source(
                module_source
            )

            logger.warning(
                "Found source with a pinned sha. Will now checkout "
                f"module: {module_source}"
            )

            # critical step
            if not self.check_module_clean(source):
                raise GitException(
                    f"One of the files at {relative_module_path} "
                    f"is not committed and we "
                    f"are trying to load that directory from git "
                    f"history due to a pinned step in the pipeline. "
                    f"Please commit the file and then run the "
                    f"pipeline."
                )

            # Check out the directory at that sha
            self.checkout(sha_or_branch=pin, directory=relative_module_path)

            # After this point, all exceptions will first undo the above
            try:
                class_ = import_class_by_path(source)
                self.reset(relative_module_path)
                self.checkout(directory=relative_module_path)
            except Exception as e:
                self.reset(relative_module_path)
                self.checkout(directory=relative_module_path)
                raise GitException(
                    f"A git exception occured when checking out repository "
                    f"from git history. Resetting repository to original "
                    f"state. Original exception: {e}"
                )

        elif "@" in source and is_standard:
            logger.debug(f"Default {APP_NAME} class used. Loading directly.")
            # TODO: [LOW] Check if ZenML version is installed before loading.
            class_ = import_class_by_path(source)
        else:
            logger.debug(
                "Unpinned step found with no git sha. Attempting to "
                "load class from current repository state."
            )
            class_ = import_class_by_path(source)

        return class_

    def resolve_class(self, class_: Type) -> Text:
        """
        Resolves
        Args:
            class_: A Python Class reference.

        Returns: source_path e.g. this.module.Class[@pin].
        """
        initial_source = class_.__module__ + "." + class_.__name__
        if is_standard_source(initial_source):
            return resolve_standard_source(initial_source)

        # Get the full module path relative to the repository
        file_path = inspect.getfile(class_)
        module_source = get_module_source_from_file_path(file_path)

        class_source = module_source + "." + class_.__name__
        return self.resolve_class_source(class_source)



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from ml_metadata.proto import metadata_store_pb2

from zenml.core.component_factory import metadata_store_factory
from zenml.enums import MLMetadataTypes
from zenml.metadata.base_metadata_store import BaseMetadataStore


@metadata_store_factory.register(MLMetadataTypes.mock)
class MockMetadataStore(BaseMetadataStore):
    """Mock metadata store."""

    def get_tfx_metadata_config(self):
        """Return tfx metadata config for mock metadata store."""
        config = metadata_store_pb2.ConnectionConfig()
        config.fake_database.SetInParent()  # Sets an empty fake
        return config



from tfx.dsl.components.common.importer import Importer
from tfx.orchestration import pipeline as tfx_pipeline
from tfx.orchestration.local.local_dag_runner import LocalDagRunner

from zenml.core.component_factory import orchestrator_store_factory
from zenml.enums import OrchestratorTypes
from zenml.orchestrators.base_orchestrator import BaseOrchestrator


@orchestrator_store_factory.register(OrchestratorTypes.local)
class LocalOrchestrator(BaseOrchestrator):
    def run(self, zenml_pipeline, **pipeline_args):
        runner = LocalDagRunner()

        # Resolve the importers for external artifact inputs
        importers = {}
        for name, artifact in zenml_pipeline.inputs.items():
            importers[name] = Importer(
                source_uri=artifact.uri, artifact_type=artifact.type
            ).with_id(name)

        import_artifacts = {
            n: i.outputs["result"] for n, i in importers.items()
        }

        # Establish the connections between the components
        zenml_pipeline.connect(**import_artifacts, **zenml_pipeline.steps)

        # Create the final step list and the corresponding pipeline
        steps = list(importers.values()) + [
            s.component for s in zenml_pipeline.steps.values()
        ]

        artifact_store = zenml_pipeline.stack.artifact_store
        metadata_store = zenml_pipeline.stack.metadata_store

        created_pipeline = tfx_pipeline.Pipeline(
            pipeline_name="pipeline_name",
            components=steps,
            pipeline_root=artifact_store.path,
            metadata_connection_config=metadata_store.get_tfx_metadata_config(),
            enable_cache=False,
        )
        return runner.run(created_pipeline)



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

from zenml.materializers.base_materializer import BaseMaterializer
from zenml.utils import yaml_utils

DEFAULT_FILENAME = "data.json"


class JsonMaterializer(BaseMaterializer):
    """Read data to and from JSON."""

    TYPE_NAME = "json"

    def read_file(self, filename=None):
        """Read JSON from filename."""
        filepath = os.path.join(
            self.artifact.uri,
            filename if filename is not None else DEFAULT_FILENAME,
        )
        return yaml_utils.read_json(filepath)

    def write_file(self, data, filename=None):
        """Write JSON to filename."""
        filepath = os.path.join(
            self.artifact.uri,
            filename if filename is not None else DEFAULT_FILENAME,
        )

        yaml_utils.write_json(filepath, data)



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Text

import click
import git

from zenml.cli.cli import cli
from zenml.cli.utils import confirmation
from zenml.core.repo import Repository
from zenml.utils.analytics_utils import INITIALIZE, track


@cli.command("init", help="Initialize zenml on a given path.")
@click.option("--repo_path", type=click.Path(exists=True))
@click.option("--pipelines_dir", type=click.Path(exists=True))
@click.option("--analytics_opt_in", "-a", type=click.BOOL)
@track(event=INITIALIZE)
def init(
    repo_path: Text, pipelines_dir: Text = None, analytics_opt_in: bool = True
):
    """Initialize ZenML on given path.

    Args:
      repo_path: Text:
      pipelines_dir: Text:  (Default value = None)
      analytics_opt_in: bool:  (Default value = True)

    Raises:
        InvalidGitRepositoryError: If repo is not a git repo.
        AssertionError
    """
    if repo_path is None:
        repo_path = os.getcwd()

    click.echo(f"Initializing at {repo_path}")

    try:
        Repository.init_repo(
            repo_path=repo_path,
            analytics_opt_in=False,
        )
        click.echo(f"ZenML repo initialized at {repo_path}")
    except git.InvalidGitRepositoryError:
        click.echo(
            f"{repo_path} is not a valid git repository! Please "
            f"initialize ZenML within a git repository using "
            f"`git init `"
        )
    except AssertionError as e:
        click.echo(f"{e}")


@cli.command("clean")
@click.option("--yes", "-y", type=click.BOOL, default=False)
def clean(yes: bool = False):
    """Clean everything in repository.

    Args:
      yes: bool:  (Default value = False)

    Returns:

    """
    if not yes:
        _ = confirmation(
            "This will completely delete all pipelines, their associated "
            "artifacts and metadata ever created in this ZenML repository. "
            "Are you sure you want to proceed?"
        )

    click.echo("Not implemented for this version")
    # if confirm:
    #     repo.clean()



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import datetime

from tfx.dsl.components.common.importer import Importer
from tfx.orchestration import pipeline as tfx_pipeline

from zenml.core.component_factory import orchestrator_store_factory
from zenml.enums import OrchestratorTypes
from zenml.orchestrators.airflow.airflow_dag_runner import (
    AirflowDagRunner,
    AirflowPipelineConfig,
)
from zenml.orchestrators.base_orchestrator import BaseOrchestrator


@orchestrator_store_factory.register(OrchestratorTypes.airflow)
class AirflowOrchestrator(BaseOrchestrator):
    def run(self, zenml_pipeline):
        # TODO: [LOW] remove or modify the configuration here
        _airflow_config = {
            "schedule_interval": None,
            "start_date": datetime.datetime(2019, 1, 1),
        }

        runner = AirflowDagRunner(AirflowPipelineConfig(_airflow_config))

        # Resolve the importers for external artifact inputs
        importers = {}
        for name, artifact in zenml_pipeline.inputs.items():
            importers[name] = Importer(
                source_uri=artifact.uri, artifact_type=artifact.type
            ).with_id(name)

        import_artifacts = {
            n: i.outputs["result"] for n, i in importers.items()
        }

        # Establish the connections between the components
        zenml_pipeline.connect(**import_artifacts, **zenml_pipeline.steps)

        # Create the final step list and the corresponding pipeline
        steps = list(importers.values()) + [
            s.get_component() for s in zenml_pipeline.steps.values()
        ]

        artifact_store = zenml_pipeline.stack.artifact_store
        metadata_store = zenml_pipeline.stack.metadata_store

        created_pipeline = tfx_pipeline.Pipeline(
            pipeline_name="pipeline_name",
            components=steps,
            pipeline_root=artifact_store.path,
            metadata_connection_config=metadata_store.get_tfx_metadata_config(),
            enable_cache=False,
        )

        return runner.run(created_pipeline)



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML specific exception definitions"""

from typing import Text


class InitializationException(Exception):
    """Raises exception when a function is run before zenml initialization."""

    def __init__(
        self, message="ZenML config is none. Did you do `zenml init`?"
    ):
        super().__init__(message)


class EmptyDatasourceException(Exception):
    """Raises exception when a datasource data is accessed without running
    an associated pipeline."""

    def __init__(
        self,
        message="This datasource has not been used in any "
        "pipelines, therefore the associated data has no "
        "versions. Please use this datasouce in any ZenML "
        "pipeline with `pipeline.add_datasource("
        "datasource)`",
    ):
        super().__init__(message)


class DoesNotExistException(Exception):
    """Raises exception when the `name` does not exist in the system but an
    action is being done that requires it to be present."""

    def __init__(
        self,
        name: Text = "",
        reason: Text = "",
        message="{} does not exist! This might be due to: {}",
    ):
        super().__init__(message.format(name, reason))


class AlreadyExistsException(Exception):
    """Raises exception when the `name` already exist in the system but an
    action is trying to create a resource with the same name."""

    def __init__(
        self, message: Text = None, name: Text = "", resource_type: Text = ""
    ):
        if message is None:
            message = f"{resource_type} `{name}` already exists!"
        super().__init__(message)


class PipelineNotSucceededException(Exception):
    """Raises exception when trying to fetch artifacts from a not succeeded
    pipeline."""

    def __init__(
        self,
        name: Text = "",
        message: Text = "{} is not yet completed successfully.",
    ):
        super().__init__(message.format(name))


class GitException(Exception):
    """Raises exception when a problem occurs in git resolution."""

    def __init__(
        self,
        message: Text = "There is a problem with git resolution. "
        "Please make sure that all relevant files "
        "are committed.",
    ):
        super().__init__(message)



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text

from tfx.orchestration import metadata

from zenml.core.component_factory import metadata_store_factory
from zenml.enums import MLMetadataTypes
from zenml.metadata.base_metadata_store import BaseMetadataStore


@metadata_store_factory.register(MLMetadataTypes.mysql)
class MySQLMetadataStore(BaseMetadataStore):
    """MySQL backend for ZenML metadata store."""

    host: Text
    port: int
    database: Text
    username: Text
    password: Text

    def get_tfx_metadata_config(self):
        """Return tfx metadata config for mysql metadata store."""
        return metadata.mysql_metadata_connection_config(
            host=self.host,
            port=self.port,
            database=self.database,
            username=self.username,
            password=self.password,
        )



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import types
from typing import Text, Type

from zenml.steps.base_step import BaseStep
from zenml.utils.exceptions import StepInterfaceError


def step(name: Text = None):
    """Outer decorator function for the creation of a ZenML step

    In order to be able work with parameters such as `name`, it features a
    nested decorator structure.

    Args:
        name (required) the given name for the step.

    Returns:
        the inner decorator which creates the step class based on the
        ZenML BaseStep
    """

    if not isinstance(name, str):
        raise StepInterfaceError("Please give your step a unique name!")

    def inner_decorator(func: types.FunctionType) -> Type:
        """Inner decorator function for the creation of a ZenML Step

        Args:
          func: types.FunctionType, this function will be used as the
            "process" method of the generated Step

        Returns:
            The class of a newly generated ZenML Step.
        """
        step_class = type(
            name if name else func.__name__,
            (BaseStep,),
            {"process": staticmethod(func)},
        )
        return step_class

    return inner_decorator



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from pathlib import Path

import apache_beam as beam

from zenml.logger import get_logger
from zenml.materializers.base_materializer import BaseMaterializer
from zenml.utils import path_utils

logger = get_logger(__name__)

DEFAULT_FILENAME = "data.csv"


class BeamMaterializer(BaseMaterializer):
    """Read to and from Beam artifacts."""

    TYPE_NAME = "beam"

    def read_text(self, pipeline, read_header: bool = True):
        """Read from text"""
        pipeline = pipeline | "ReadText" >> beam.io.ReadFromText(
            file_pattern=os.path.join(self.artifact.uri, "*"),
            skip_header_lines=True,
        )

        if read_header:
            wildcard_qualifier = "*"
            file_pattern = os.path.join(self.artifact.uri, wildcard_qualifier)

            csv_files = path_utils.list_dir(self.artifact.uri)
            if not csv_files:
                raise RuntimeError(
                    "Split pattern {} does not match any files.".format(
                        file_pattern
                    )
                )

            # weed out bad file exts with this logic
            allowed_file_exts = [".csv", ".txt"]  # ".dat"
            csv_files = [
                uri
                for uri in csv_files
                if Path(uri).suffix in allowed_file_exts
            ]

            logger.info(f"Matched {len(csv_files)}: {csv_files}")

            # Always use header from file
            logger.info(f"Using header from file: {csv_files[0]}.")
            header = ",".join(path_utils.load_csv_header(csv_files[0]))

            return header, pipeline
        else:
            return pipeline

    def write_text(self, pipeline, shard_name_template=None, header=None):
        """Write from text"""
        return pipeline | beam.io.WriteToText(
            os.path.join(self.artifact.uri, DEFAULT_FILENAME),
            shard_name_template=shard_name_template,
            header=header,
            num_shards=3,
        )



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from zenml.materializers.materializer_factory import MaterializerFactory


class BaseMaterializerMeta(type):
    def __new__(mcs, name, bases, dct):
        """ """

        cls = super().__new__(mcs, name, bases, dct)
        if name != "BaseMaterializer":
            assert cls.TYPE_NAME != "base", (
                f"You have used the name `base` as a TYPE_NAME "
                f"for your class {name}. When you are defining a new "
                f"materializer, please make sure that you give it a "
                f"TYPE_NAME other than `base`."
            )
            MaterializerFactory.register_type(cls.TYPE_NAME, cls)
        return cls


class BaseMaterializer(metaclass=BaseMaterializerMeta):
    """Base Materializer to realize artifact data."""

    TYPE_NAME = "base"

    def __init__(self, artifact):
        self.artifact = artifact



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

"""
The collection of utility functions/classes are inspired by their original
implementation of the Tensorflow Extended team, which can be found here:

https://github.com/tensorflow/tfx/blob/master/tfx/dsl/component/experimental/decorators.py

This version is heavily adjusted to work with the Pipeline-Step paradigm which
is proposed by ZenML.
"""

from __future__ import absolute_import, division, print_function

import inspect
import json
import sys
from typing import Any, Callable, Dict, List, Optional, Text, Type

import pydantic
from pydantic import create_model
from tfx import types as tfx_types
from tfx.dsl.component.experimental.decorators import _SimpleComponent
from tfx.dsl.components.base.base_executor import BaseExecutor
from tfx.dsl.components.base.executor_spec import ExecutorClassSpec
from tfx.types import component_spec
from tfx.types.channel import Channel
from tfx.utils import json_utils

from zenml.annotations import Input, Output
from zenml.artifacts.base_artifact import BaseArtifact
from zenml.utils.exceptions import StepInterfaceError


class _PropertyDictWrapper(json_utils.Jsonable):
    """Helper class to wrap inputs/outputs from TFX nodes.
    Currently, this class is read-only (setting properties is not implemented).
    Internal class: no backwards compatibility guarantees.
    Code Credit: https://github.com/tensorflow/tfx/blob/51946061ae3be656f1718a3d62cd47228b89b8f4/tfx/types/node_common.py
    """

    def __init__(
        self,
        data: Dict[str, Channel],
        compat_aliases: Optional[Dict[str, str]] = None,
    ):
        self._data = data
        self._compat_aliases = compat_aliases or {}

    def __iter__(self):
        yield from self._data

    def __getitem__(self, key):
        if key in self._compat_aliases:
            key = self._compat_aliases[key]
        return self._data[key]

    def __getattr__(self, key):
        if key in self._compat_aliases:
            key = self._compat_aliases[key]
        try:
            return self._data[key]
        except KeyError:
            raise AttributeError

    def __repr__(self):
        return repr(self._data)

    def get_all(self) -> Dict[str, Channel]:
        return self._data

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()


class _ZenMLSimpleComponent(_SimpleComponent):
    """Simple ZenML TFX component with outputs overridden."""

    @property
    def outputs(self) -> _PropertyDictWrapper:
        return _PropertyDictWrapper(self.spec.outputs)


class _FunctionExecutor(BaseExecutor):
    """ """

    _FUNCTION = staticmethod(lambda: None)

    def Do(
        self,
        input_dict: Dict[Text, List[tfx_types.Artifact]],
        output_dict: Dict[Text, List[tfx_types.Artifact]],
        exec_properties: Dict[Text, Any],
    ):
        function_args = {}
        for name, artifact in input_dict.items():
            if len(artifact) == 1:
                function_args[name] = artifact[0]
            else:
                raise ValueError(
                    (
                        "Expected input %r to %s to be a singleton "
                        "ValueArtifact channel (got %s instead)."
                    )
                    % (name, self, artifact)
                )

        return_artifact = output_dict.pop("return_output", None)
        for name, artifact in output_dict.items():
            if len(artifact) == 1:
                function_args[name] = artifact[0]
            else:
                raise ValueError(
                    (
                        "Expected output %r to %s to be a singleton "
                        "ValueArtifact channel (got %s instead)."
                    )
                    % (name, self, artifact)
                )

        ####
        spec = inspect.getfullargspec(self._FUNCTION)

        args = spec.args

        inputs_to_take_care_of = []
        param_spec = {}
        for arg in args:
            arg_type = spec.annotations.get(arg, None)
            if isinstance(arg_type, Input):
                if not issubclass(arg_type.type, BaseArtifact):
                    inputs_to_take_care_of.append(arg)
            elif isinstance(arg_type, Output):
                pass
            else:
                param_spec.update({arg: arg_type})

        new_exec = {k: json.loads(v) for k, v in exec_properties.items()}
        pydantic_c: Type[pydantic.BaseModel] = create_model(
            "params", **{k: (v, ...) for k, v in param_spec.items()}
        )
        function_args.update(pydantic_c.parse_obj(new_exec).dict())

        returns = self._FUNCTION(**function_args)

        if return_artifact is not None:
            artifact = return_artifact[0]
            if returns is not None:
                artifact.materializers.json.write_file(returns)
            else:
                raise StepInterfaceError()

        if returns is not None and return_artifact is None:
            raise StepInterfaceError()


def generate_component(step) -> Callable[..., Any]:
    spec_inputs, spec_outputs, spec_params = {}, {}, {}
    for key, artifact_type in step.INPUT_SPEC.items():
        spec_inputs[key] = component_spec.ChannelParameter(type=artifact_type)
    for key, artifact_type in step.OUTPUT_SPEC.items():
        spec_outputs[key] = component_spec.ChannelParameter(type=artifact_type)
    for key, prim_type in step.PARAM_SPEC.items():
        spec_params[key] = component_spec.ExecutionParameter(type=str)

    component_spec_class = type(
        "%s_Spec" % step.__class__.__name__,
        (tfx_types.ComponentSpec,),
        {
            "INPUTS": spec_inputs,
            "OUTPUTS": spec_outputs,
            "PARAMETERS": spec_params,
        },
    )

    executor_class = type(
        "%s_Executor" % step.__class__.__name__,
        (_FunctionExecutor,),
        {
            "_FUNCTION": staticmethod(step.process),
            "__module__": step.__module__,
        },
    )

    module = sys.modules[step.__module__]
    setattr(module, "%s_Executor" % step.__class__.__name__, executor_class)

    executor_spec_instance = ExecutorClassSpec(executor_class=executor_class)

    return type(
        step.__class__.__name__,
        (_ZenMLSimpleComponent,),
        {
            "SPEC_CLASS": component_spec_class,
            "EXECUTOR_SPEC": executor_spec_instance,
            "__module__": step.__module__,
        },
    )



# #  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
# #
# #  Licensed under the Apache License, Version 2.0 (the "License");
# #  you may not use this file except in compliance with the License.
# #  You may obtain a copy of the License at:
# #
# #       http://www.apache.org/licenses/LICENSE-2.0
# #
# #  Unless required by applicable law or agreed to in writing, software
# #  distributed under the License is distributed on an "AS IS" BASIS,
# #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# #  or implied. See the License for the specific language governing
# #  permissions and limitations under the License.
#
# import os
# import random
# from pathlib import Path
# from typing import Text
#
# import pytest
#
# import zenml
# from zenml.core.repo import Repository
# from zenml.pipelines import BasePipeline
# from zenml.utils import yaml_utils
# from zenml import __version__
#
# # Nicholas a way to get to the root
# ZENML_ROOT = str(Path(zenml.__path__[0]).parent)
# TEST_ROOT = os.path.join(ZENML_ROOT, "tests")
#
#
# def test_repo_double_init():
#     # explicitly constructing another repository should fail
#     with pytest.raises(Exception):
#         _ = Repository()
#
#
# def test_get_datasources(repo):
#     ds_list = repo.get_datasources()
#
#     assert "my_csv_datasource" in [x.name for x in ds_list]
#
#
# def test_get_datasource_by_name(repo):
#     assert repo.get_datasource_by_name("my_csv_datasource")
#
#     fake_ds = repo.get_datasource_by_name("ds_123")
#
#     assert fake_ds is None
#
#
# def test_get_datasource_names(repo):
#     # TODO [LOW]: Automatically expand when new datasource tests are added!
#     test_ds_names = [
#         "my_csv_datasource",
#         "image_ds_local",
#         "image_ds_gcp",
#         "json_ds",
#     ]
#
#     ds_names = repo.get_datasource_names()
#
#     assert set(test_ds_names) <= set(ds_names)
#
#
# def test_get_pipeline_file_paths(repo, monkeypatch):
#     mock_paths = ["pipeline_1.yaml", "pipeline_2.yaml", "awjfof.txt"]
#
#     def mock_list_dir(dir_path: Text, only_file_names: bool = False):
#         # add a corrupted file into the pipelines
#         return mock_paths
#
#     monkeypatch.setattr("zenml.utils.path_utils.list_dir", mock_list_dir)
#
#     paths = repo.get_pipeline_file_paths(only_file_names=True)
#
#     assert paths == mock_paths[:-1]
#
#
# def test_get_pipeline_names(repo):
#     # TODO: This has to be made dynamic once more pipelines come
#     real_p_names = sorted(["csvtest{0}".format(i) for i in range(1, 6)])
#
#     found_p_names = sorted(repo.get_pipeline_names())
#
#     assert set(real_p_names) <= set(found_p_names)
#
#
# def test_get_pipelines(repo):
#     p_names = sorted(repo.get_pipeline_names())
#
#     pipelines = repo.get_pipelines()
#
#     pipelines = sorted(pipelines, key=lambda p: p.name)
#
#     assert all(p.name == name for p, name in zip(pipelines, p_names))
#
#
# def test_get_pipelines_by_type(repo):
#     pipelines = repo.get_pipelines_by_type(type_filter=["training"])
#
#     pipelines_2 = repo.get_pipelines_by_type(type_filter=["base"])
#
#     assert len(pipelines) == 5
#
#     assert not pipelines_2
#
#
# def test_get_pipeline_by_name(repo, equal_pipelines):
#     p_names = repo.get_pipeline_names()
#
#     random_name = random.choice(p_names)
#     cfg_list = [y for y in repo.get_pipeline_file_paths() if random_name in y]
#
#     cfg = yaml_utils.read_yaml(cfg_list[0])
#
#     p1 = repo.get_pipeline_by_name(random_name)
#
#     p2 = BasePipeline.from_config(cfg)
#
#     assert equal_pipelines(p1, p2, loaded=True)
#
#
# def test_get_step_versions(repo):
#     step_versions = repo.get_step_versions()
#
#     # TODO: Make this less hardcoded
#     steps_used = [
#         "zenml.steps.preprocesser.standard_preprocesser.standard_preprocesser.StandardPreprocesser",
#         "zenml.steps.split.categorical_domain_split_step.CategoricalDomainSplit",
#         "zenml.steps.trainer.tensorflow_trainers.tf_ff_trainer.FeedForwardTrainer",
#     ]
#
#     current_version = "zenml_" + str(__version__)
#
#     assert set(steps_used) >= set(step_versions.keys())
#     assert all(current_version in s for s in step_versions.values())
#
#
# def test_get_step_by_version(repo):
#     # TODO: Make this less hardcoded
#     steps_used = [
#         "zenml.steps.preprocesser.standard_preprocesser.standard_preprocesser.StandardPreprocesser",
#         "zenml.steps.split.categorical_domain_split_step.CategoricalDomainSplit",
#         "zenml.steps.trainer.tensorflow_trainers.tf_ff_trainer.FeedForwardTrainer",
#     ]
#
#     random_step = random.choice(steps_used)
#
#     current_version = "zenml_" + str(__version__)
#
#     bogus_version = "asdfghjklöä"
#
#     assert repo.get_step_by_version(random_step, current_version)
#     assert repo.get_step_by_version(random_step, bogus_version) is None
#
#
# def test_get_step_versions_by_type(repo):
#     # TODO: Make this less hardcoded
#     steps_used = [
#         "zenml.steps.preprocesser.standard_preprocesser.standard_preprocesser.StandardPreprocesser",
#         "zenml.steps.split.categorical_domain_split_step.CategoricalDomainSplit",
#         "zenml.steps.trainer.tensorflow_trainers.tf_ff_trainer.FeedForwardTrainer",
#     ]
#
#     random_step = random.choice(steps_used)
#
#     current_version = "zenml_" + str(__version__)
#
#     bogus_step = "asdfghjklöä"
#
#     step_versions = repo.get_step_versions_by_type(random_step)
#
#     assert step_versions == {current_version}
#
#     assert repo.get_step_versions_by_type(bogus_step) is None



#  Copyright (c) ZenML GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

from zenml.logger import init_logging

init_logging()
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

from .base import *  # noqa
from .config import *  # noqa

# from .example import *  # noqa
from .stack import *  # noqa
from .version import *  # noqa
